@@ -3,12 +3,12 @@
 from pythran.analyses.global_declarations import GlobalDeclarations
 from pythran.intrinsic import Intrinsic, Class, UnboundValue
 from pythran.passmanager import ModuleAnalysis
-from pythran.syntax import PythranSyntaxError
 from pythran.tables import functions, methods, MODULES
 from pythran.unparse import Unparser
 from pythran.conversion import demangle
 import pythran.metadata as md
 from pythran.utils import isnum
+from pythran.syntax import PythranSyntaxError
 
 import gast as ast
 from copy import deepcopy
@@ -18,32 +18,42 @@ import io
 
 IntrinsicAliases = dict()
 
+class UnboundIdentifierError(RuntimeError):
+    pass
+
+
+class DictLayer(object):
+    def __init__(self, base, layer=None):
+        self.base = base
+        self.layer = dict() if layer is None else layer
+
+    def __getitem__(self, key):
+        return (self.layer if key in self.layer else self.base)[key]
+
+    def __contains__(self, key):
+        return (key in self.layer) or (key in self.base)
+
+    def __setitem__(self, key, value):
+        self.layer[key] = value
+
+    def get(self, key, default=None):
+        return self.layer.get(key, self.base.get(key, default))
+
 
 class ContainerOf(object):
     '''
     Represents a container of something
 
     We just know that if indexed by the integer value `index',
-    we get `containee'
+    we get `containees'
     '''
     UnknownIndex = float('nan')
 
-    __slots__ = 'index', 'containee'
+    __slots__ = 'index', 'containees'
 
-    cache = {}
-
-    def __new__(cls, *args, **kwargs):
-        # cache the creation of new objects, so that same keys give same id
-        # thus great hashing
-        key = tuple(args), tuple(kwargs.items())
-        if key not in ContainerOf.cache:
-            new_obj = super(ContainerOf, cls).__new__(cls)
-            ContainerOf.cache[key] = new_obj
-        return ContainerOf.cache[key]
-
-    def __init__(self, containee, index=UnknownIndex):
+    def __init__(self, containees, index=UnknownIndex):
         self.index = index
-        self.containee = containee
+        self.containees = containees
 
 
 def save_intrinsic_alias(module):
@@ -74,14 +84,19 @@ class Aliases(ModuleAnalysis):
     def __init__(self):
         self.result = dict()
         self.aliases = None
-        ContainerOf.cache.clear()
         super(Aliases, self).__init__(GlobalDeclarations)
 
     @staticmethod
     def dump(result, filter=None):
         def pp(n):
             output = io.StringIO()
-            Unparser(n, output)
+            if isinstance(n, ContainerOf):
+                if n.index == n.index:  # always valid except for UnknownIndex
+                    output.write('[{}]='.format(n.index))
+                containees = sorted(map(pp, n.containees))
+                output.write(', '.join(map("|{}|".format, containees)))
+            else:
+                Unparser(n, output)
             return output.getvalue().strip()
 
         if isinstance(result, dict):
@@ -183,7 +198,7 @@ class Aliases(ModuleAnalysis):
         >>> module = ast.parse('def foo(a, b): return {0: a, 1: b}')
         >>> result = pm.gather(Aliases, module)
         >>> Aliases.dump(result, filter=ast.Dict)
-        {0: a, 1: b} => ['|a|', '|b|']
+        {0: a, 1: b} => ['|a|, |b|']
 
         where the |id| notation means something that may contain ``id``.
         '''
@@ -192,10 +207,11 @@ class Aliases(ModuleAnalysis):
             for key, val in zip(node.keys, node.values):
                 self.visit(key)  # res ignored, just to fill self.aliases
                 elt_aliases = self.visit(val)
-                elts_aliases.update(map(ContainerOf, elt_aliases))
+                elts_aliases.update(elt_aliases)
+            aliases = {ContainerOf(elts_aliases)}
         else:
-            elts_aliases = None
-        return self.add(node, elts_aliases)
+            aliases = None
+        return self.add(node, aliases)
 
     def visit_Set(self, node):
         '''
@@ -206,14 +222,14 @@ class Aliases(ModuleAnalysis):
         >>> module = ast.parse('def foo(a, b): return {a, b}')
         >>> result = pm.gather(Aliases, module)
         >>> Aliases.dump(result, filter=ast.Set)
-        {a, b} => ['|a|', '|b|']
+        {a, b} => ['|a|, |b|']
 
         where the |id| notation means something that may contain ``id``.
         '''
         if node.elts:
-            elts_aliases = {ContainerOf(alias)
+            elts_aliases = {ContainerOf({alias
                             for elt in node.elts
-                            for alias in self.visit(elt)}
+                            for alias in self.visit(elt)})}
         else:
             elts_aliases = None
         return self.add(node, elts_aliases)
@@ -268,15 +284,13 @@ class Aliases(ModuleAnalysis):
             arg_aliases = [self.result[arg] or {arg} for arg in args]
             return_aliases = set()
             for args_combination in product(*arg_aliases):
-                return_aliases.update(
-                    func.return_alias(args_combination))
-            return {expand_subscript(ra) for ra in return_aliases}
-
-        def expand_subscript(node):
-            if isinstance(node, ast.Subscript):
-                if isinstance(node.value, ContainerOf):
-                    return node.value.containee
-            return node
+                for ra in func.return_alias(args_combination):
+                    if isinstance(ra, ast.Subscript):
+                        if isinstance(ra.value, ContainerOf):
+                            return_aliases.update(ra.value.containees)
+                            continue
+                    return_aliases.add(ra)
+            return return_aliases
 
         def full_args(func, call):
             args = call.args
@@ -311,7 +325,9 @@ class Aliases(ModuleAnalysis):
                     aliases.update(interprocedural_aliases(func_alias, args))
                 else:
                     pass  # better thing to do ?
-        [self.add(a) for a in aliases if a not in self.result]
+        for a in aliases:
+            if a not in self.result:
+                self.add(a)
         return aliases or self.get_unbound_value_set()
 
     def visit_Call(self, node):
@@ -440,7 +456,7 @@ class Aliases(ModuleAnalysis):
                         if node.slice.value != alias.index:
                             continue
                     # FIXME: what if the index is a slice variable...
-                    aliases.add(alias.containee)
+                    aliases.update(alias.containees)
                 elif isinstance(getattr(alias, 'ctx', None), (ast.Param,
                                                               ast.Store)):
                     aliases.add(ast.Subscript(alias, node.slice, node.ctx))
@@ -457,11 +473,7 @@ class Aliases(ModuleAnalysis):
 
     def visit_Name(self, node):
         if node.id not in self.aliases:
-            err = ("identifier {0} unknown, either because "
-                   "it is an unsupported intrinsic, "
-                   "the input code is faulty, "
-                   "or... pythran is buggy.")
-            raise PythranSyntaxError(err.format(node.id), node)
+            raise UnboundIdentifierError
         return self.add(node, self.aliases[node.id])
 
     def visit_Tuple(self, node):
@@ -473,20 +485,19 @@ class Aliases(ModuleAnalysis):
         >>> module = ast.parse('def foo(a, b): return a, b')
         >>> result = pm.gather(Aliases, module)
         >>> Aliases.dump(result, filter=ast.Tuple)
-        (a, b) => ['|[0]=a|', '|[1]=b|']
+        (a, b) => ['[0]=|a|', '[1]=|b|']
 
         where the |[i]=id| notation means something that
         may contain ``id`` at index ``i``.
         '''
         if node.elts:
-            elts_aliases = set()
+            aliases = set()
             for i, elt in enumerate(node.elts):
                 elt_aliases = self.visit(elt)
-                elts_aliases.update(ContainerOf(alias, i)
-                                    for alias in elt_aliases)
+                aliases.add(ContainerOf(elt_aliases, i))
         else:
-            elts_aliases = None
-        return self.add(node, elts_aliases)
+            aliases = None
+        return self.add(node, aliases)
 
     visit_List = visit_Set
 
@@ -558,11 +569,10 @@ class Aliases(ModuleAnalysis):
                 if isinstance(exp, (ast.Constant, Intrinsic, ast.FunctionDef)):
                     return lambda _: {exp}
                 elif isinstance(exp, ContainerOf):
-                    pcontainee = parametrize(exp.containee)
                     index = exp.index
                     return lambda args: {
-                        ContainerOf(pc, index)
-                        for pc in pcontainee(args)
+                        ContainerOf({pc for containee in exp.containees for pc
+                                     in parametrize(containee)(args)}, index)
                     }
                 elif isinstance(exp, ast.Name):
                     try:
@@ -651,13 +661,14 @@ class Aliases(ModuleAnalysis):
         ...         {i}""")
         >>> result = pm.gather(Aliases, module)
         >>> Aliases.dump(result, filter=ast.Set)
-        {i} => ['|a|', '|b|']
+        {i} => ['|a|, |b|']
         '''
 
         iter_aliases = self.visit(node.iter)
         if all(isinstance(x, ContainerOf) for x in iter_aliases):
-            target_aliases = {iter_alias.containee for iter_alias in
-                              iter_aliases}
+            target_aliases = set().union(*[iter_alias.containees
+                                           for iter_alias in
+                              iter_aliases])
         else:
             target_aliases = {node.target}
 
@@ -702,7 +713,7 @@ class Aliases(ModuleAnalysis):
         >>> module = ast.parse(fun)
         >>> result = pm.gather(Aliases, module)
         >>> Aliases.dump(result, filter=ast.Set)
-        {c} => ['|a|', '|b|']
+        {c} => ['|a|, |b|']
         '''
 
         md.visit(self, node)
@@ -711,44 +722,55 @@ class Aliases(ModuleAnalysis):
 
         # first try the true branch
         try:
-            tmp = self.aliases.copy()
+            self.aliases = DictLayer(self.aliases)
             for stmt in node.body:
                 self.visit(stmt)
             true_aliases = self.aliases
-            self.aliases = tmp
-        except PythranSyntaxError:
+        except UnboundIdentifierError:
             pass
+        finally:
+            self.aliases = self.aliases.base
 
         # then try the false branch
         try:
+            self.aliases = DictLayer(self.aliases)
             for stmt in node.orelse:
                 self.visit(stmt)
             false_aliases = self.aliases
-        except PythranSyntaxError:
+        except UnboundIdentifierError:
             pass
+        finally:
+            self.aliases = self.aliases.base
 
         if true_aliases and not false_aliases:
-            self.aliases = true_aliases
+            self.aliases = DictLayer(true_aliases)
             for stmt in node.orelse:
                 self.visit(stmt)
             false_aliases = self.aliases
+            self.aliases = self.aliases.base
 
         if false_aliases and not true_aliases:
-            self.aliases = false_aliases
+            self.aliases = DictLayer(false_aliases)
             for stmt in node.body:
                 self.visit(stmt)
             true_aliases = self.aliases
+            self.aliases = self.aliases.base
 
         # merge the results from true and false branches
-        if false_aliases and true_aliases:
-            for k, v in true_aliases.items():
-                if k in self.aliases:
-                    self.aliases[k] = self.aliases[k].union(v)
-                else:
-                    assert isinstance(v, (frozenset, set))
-                    self.aliases[k] = v
-        elif true_aliases:
-            self.aliases = true_aliases
+        for k, v in true_aliases.layer.items():
+            if k in false_aliases.layer:
+                self.aliases[k] = v.union(false_aliases[k])
+            elif k in self.aliases:
+                self.aliases[k] = v.union(self.aliases[k])
+            else:
+                self.aliases[k] = v
+        for k, v in false_aliases.layer.items():
+            if k in true_aliases.layer:
+                pass  # already done
+            elif k in self.aliases:
+                self.aliases[k] = v.union(self.aliases[k])
+            else:
+                self.aliases[k] = v
 
     def visit_ExceptHandler(self, node):
         if node.name:
